import torch
import torchph.nn.slayer
import torchvision.models

class Slayer(torch.nn.Module):
    def __init__(self, n_elements: int):
        super().__init__()
        self.slayer = torchph.nn.slayer.SLayerExponential(n_elements)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        output = self.slayer(input)
        # There as a squeeze() at the end of SLayerExponential.forward()
        # that breaks everything if n_elements=1 or batch_size=1
        if len(output.shape) != 1:
            return output
        if self.slayer.n_elements == 1:
            return output.unsqueeze(1)
        return output.unsqueeze(0)


class NNBase(torch.nn.Module):
    def __init__(
        self,

        num_diagrams: int,
        skip_diagrams: bool = False,
        features_per_diagram: int = 8,

        skip_images: bool = False,
        images_n_channels: int = 3,
        images_output: int = 512,

        skip_features: bool = False,
        features_output: int = 512,

        base = torchvision.models.resnet34
    ):
        super().__init__()
        
        def make_slayer():
            n = features_per_diagram
            return torch.nn.Sequential(
                Slayer(n), torch.nn.Linear(n, n), torch.nn.BatchNorm1d(n), torch.nn.GELU(),
            )
        self.slayers_ = None if skip_diagrams else torch.nn.ModuleList([ make_slayer() for _ in range(num_diagrams) ])
        self.slayers_compressor_ = torch.nn.Linear(num_diagrams * features_per_diagram, features_output)

        self.images_ = None
        if not skip_images:
            self.images_ = base(num_classes = images_output)
            self.images_.conv1 = torch.nn.Conv2d(images_n_channels, 64, kernel_size = 7, stride = 2, padding = 3, bias = False)

        self.features_ = None if skip_features else torch.nn.LazyLinear(features_output)
    
    def forward(self, images, features, *diagrams):
        result = [ ]
        
        if self.images_ is not None:
            result.append(self.images_(images))

        if self.slayers_ is not None:
            result.append(
                self.slayers_compressor_(
                    torch.hstack([
                        self.slayers_[i]((diagrams[2 * i], diagrams[2 * i + 1], diagrams[2 * i].shape[1], len(images)))
                        for i in range(len(self.slayers_))
                    ])
                )
            )
    
        if self.features_ is not None:
            result.append(self.features_(features))

        return torch.cat(result, dim = 1)



import typing

import numpy
import sklearn.base

import cvtda.dumping
from .process_iter import process_iter

def process_iter_dump(
    transformer: sklearn.base.TransformerMixin,
    data: numpy.ndarray,
    do_fit: bool,
    dump_name: typing.Optional[str] = None,
    *args,
    **kwargs
):
    if do_fit and cvtda.dumping.dumper().has_dump(dump_name):
        # We must call fit() anyway, even if we have a dump
        transformer.fit(data, *args, **kwargs)
    function = lambda: process_iter(transformer, data, do_fit, *args, **kwargs)
    return function() if dump_name is None else cvtda.dumping.dumper().execute(function, dump_name)



import joblib
import typing
import inspect

import cvtda.logging
import cvtda.dumping

def parallel(
    function,
    iterable,
    n_jobs: int = -1,
    return_as: str = 'list',
    num_parameters: typing.Optional[str] = None
):
    def init_worker(logger, dumper):
        cvtda.logging.BaseLogger.current_logger = logger
        cvtda.dumping.BaseDumper.current_dumper = dumper

    initargs = (cvtda.logging.logger(), cvtda.dumping.dumper())
    num_parameters = num_parameters or len(inspect.signature(function).parameters)
    with joblib.parallel_backend(backend = 'loky', initializer = init_worker, initargs = initargs):
        if num_parameters == 1:
            return joblib.Parallel(return_as = return_as, n_jobs = n_jobs)(
                joblib.delayed(function)(item) for item in iterable
            )
        else:
            return joblib.Parallel(return_as = return_as, n_jobs = n_jobs)(
                joblib.delayed(function)(*item) for item in iterable
            )



import numpy

import skimage.color

import cvtda.utils
import cvtda.logging


def rgb2hsv(images: numpy.ndarray, n_jobs: int = -1) -> numpy.ndarray:
    return numpy.stack(
        cvtda.utils.parallel(
            skimage.color.rgb2hsv, cvtda.logging.logger().pbar(images, desc = "rgb2hsv"),
            n_jobs = n_jobs, return_as = 'list', num_parameters = 1
        )
    )



import typing

import torch

from .conv import make_conv

class DownsampleBlock(torch.nn.Module):
    def __init__(self, in_channels: int, out_channels: int):
        super().__init__()
        self.conv = make_conv(in_channels, out_channels)
        self.pool = torch.nn.MaxPool2d(2)
    
    def forward(self, input: torch.Tensor) -> typing.Tuple[torch.Tensor, torch.Tensor]:
        features = self.conv(input)
        output = self.pool(features)
        return features, output



import math
import typing
import dataclasses

import numpy
import joblib
import itertools
import gtda.images
import gtda.homology

import cvtda.utils
import cvtda.logging

from . import utils
import cvtda.dumping
from .interface import TopologicalExtractor
from .DiagramVectorizer import DiagramVectorizer


class FiltrationExtractor(TopologicalExtractor):
    def __init__(
        self,
        filtration_class,
        filtation_kwargs: dict,
        binarizer_threshold: float,
        diagram_settings: DiagramVectorizer.Settings,

        n_jobs: int = -1,
        return_diagrams: bool = False,
        only_get_from_dump: bool = False,
        **kwargs
    ):
        super().__init__(
            enabled = True,
            filtration_class = filtration_class,
            filtation_kwargs = filtation_kwargs,
            binarizer_threshold = binarizer_threshold,
            vectorizer_settings = diagram_settings,
            supports_rgb = False,
            n_jobs = n_jobs,
            return_diagrams = return_diagrams,
            diagram_settings = diagram_settings,
            only_get_from_dump = only_get_from_dump,
            **kwargs
        )

        self.binarizer_ = gtda.images.Binarizer(threshold = binarizer_threshold, n_jobs = self.n_jobs_)
        self.filtration_ = filtration_class(**filtation_kwargs, n_jobs = self.n_jobs_)
        self.persistence_ = None


    def get_diagrams_(self, images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None):
        cvtda.logging.logger().print(f"FiltrationExtractor: processing {dump_name}, do_fit = {do_fit}, filtration = {self.filtration_}")
        
        if do_fit and (self.persistence_ is None):
            dims = list(range(len(images.shape) - 1))
            self.persistence_ = gtda.homology.CubicalPersistence(homology_dimensions = dims, n_jobs = self.n_jobs_)
        
        bin_images = utils.process_iter(self.binarizer_, images, do_fit)
        assert bin_images.shape == images.shape

        filtrations = utils.process_iter(self.filtration_, bin_images, do_fit)
        assert filtrations.shape == images.shape

        return utils.process_iter_dump(self.persistence_, filtrations, do_fit, self.diagrams_dump_(dump_name))


class FiltrationsExtractor(cvtda.utils.FeatureExtractorBase):
    @dataclasses.dataclass(frozen = True)
    class Settings:
        binarizer_thresholds: typing.List[float] = dataclasses.field(default_factory = lambda: [ 0.25, 0.5, 0.75 ])

        height_directions: typing.Optional[typing.Iterable[typing.Tuple[float, float]]] = None
        num_radial: int = 4
        dilation: bool = False
        erosion: bool = False
        signed_distance: bool = False
        density_radiuses: typing.Iterable[int] = dataclasses.field(default_factory = lambda: [])
        
        vectorizer: DiagramVectorizer.Settings = DiagramVectorizer.Settings()

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full = Settings(
            binarizer_thresholds = [ 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9 ],
            height_directions = None,
            num_radial = 4,
            dilation = True,
            erosion = True,
            signed_distance = True,
            density_radiuses = [ 1, 3 ],
            vectorizer = DiagramVectorizer.PRESETS.full
        ),
        reduced = Settings(
            binarizer_thresholds = [ 0.25, 0.5, 0.75 ],
            height_directions = None,
            num_radial = 4,
            dilation = False,
            erosion = False,
            signed_distance = False,
            density_radiuses = [ ],
            vectorizer = DiagramVectorizer.PRESETS.reduced
        ),
        quick = Settings(
            binarizer_thresholds = [ 0.5 ],
            height_directions = [(-1, -1), (1, -1), (-1, 1), (1, 1)],
            num_radial = 2,
            dilation = False,
            erosion = False,
            signed_distance = False,
            density_radiuses = [ ],
            vectorizer = DiagramVectorizer.PRESETS.quick
        ),
    )

    def __init__(
        self,
        settings: Settings,
        n_jobs: int = -1,
        return_diagrams: bool = False,
        only_get_from_dump: bool = False,
    ):
        self.fitted_ = False
        self.n_jobs_ = n_jobs
        self.settings_ = settings
        self.return_diagrams_ = return_diagrams
        self.only_get_from_dump_ = only_get_from_dump
        self.filtration_extractors_: typing.List[typing.Tuple[FiltrationExtractor, str, str]] = []

    def feature_names(self) -> typing.List[str]:
        feature_names = []
        for extractor, _, readable_name in self.filtration_extractors_:
            feature_names.extend(self.nest_feature_names(readable_name, extractor.feature_names()))
        return feature_names

    def fit(self, images: numpy.ndarray, dump_name: typing.Optional[str] = None):
        self.fit_transform(images, dump_name)
        return self
    
    def transform(self, images: numpy.ndarray, dump_name: typing.Optional[str] = None) -> numpy.ndarray:
        assert self.fitted_ is True, 'fit() must be called before transform()'
        cvtda.logging.logger().print("Applying filtrations")
        return self.do_work_(images, do_fit = False, dump_name = dump_name)
    
    def fit_transform(self, images: numpy.ndarray, dump_name: typing.Optional[str] = None) -> numpy.ndarray:
        assert len(images.shape) >= 3, f'{len(images.shape) - 1}d images are not supported'
        cvtda.logging.logger().print("Fitting filtrations")
        
        shape = images.shape
        if (len(shape) == 4) and (shape[-1] == 3):
            shape = shape[:-1]
        self.fill_filtrations_(*shape[1:])

        result = self.do_work_(images, do_fit = True, dump_name = dump_name)
        self.fitted_ = True
        return result

    def do_work_(self, images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None) -> numpy.ndarray:
        def do_work_one(extractor: FiltrationExtractor, name, readable_name):
            extractor_dump_name = cvtda.dumping.dump_name_concat(dump_name, name)
            with cvtda.logging.DevNullLogger():
                result = utils.process_iter(extractor, images, do_fit, dump_name = extractor_dump_name)
                return ((extractor, name, readable_name), result) if do_fit else result

        parallel = cvtda.utils.parallel(do_work_one, self.filtration_extractors_, return_as = 'generator', n_jobs = self.outer_n_jobs_)
        outputs = list(cvtda.logging.logger().pbar(parallel, total = len(self.filtration_extractors_)))
        if do_fit:
            self.filtration_extractors_ = [ output[0] for output in outputs ]
            outputs = [ output[1] for output in outputs ]
        result = utils.hstack(outputs, not self.return_diagrams_)
        if not self.return_diagrams_ and not self.only_get_from_dump_:
            assert result.shape == (len(images), len(self.feature_names()))
        return result


    def fill_filtrations_(self, *shape: typing.List[int]):
        self.inner_n_jobs_ = 1
        self.do_fill_filtrations_(*shape)
        n_jobs = joblib.effective_n_jobs(self.n_jobs_)
        if len(self.filtration_extractors_) > n_jobs * 3:
            self.outer_n_jobs_ = -1
            self.inner_n_jobs_ = 1
        else:
            self.outer_n_jobs_ = math.gcd(n_jobs, len(self.filtration_extractors_))
            self.inner_n_jobs_ = n_jobs // self.outer_n_jobs_
        self.filtration_extractors_ = []
        self.do_fill_filtrations_(*shape)
    
    def do_fill_filtrations_(self, *shape: typing.List[int]):
        self.filtration_extractors_ = [ ]
        for binarizer_threshold in self.settings_.binarizer_thresholds:
            self.add_height_filtrations_(binarizer_threshold, *shape)
            self.add_radial_filtrations_(binarizer_threshold, *shape)
            self.add_dilation_filtrations_(binarizer_threshold)
            self.add_erosion_filtrations_(binarizer_threshold)
            self.add_signed_distance_filtrations_(binarizer_threshold)
            self.add_density_filtrations_(binarizer_threshold)

    def make_filtration_(self, filtration_class, filtation_kwargs: dict, binarizer_threshold: float):
        return FiltrationExtractor(
            filtration_class,
            filtation_kwargs,
            binarizer_threshold,
            self.settings_.vectorizer,
            n_jobs = self.inner_n_jobs_,
            return_diagrams = self.return_diagrams_,
            only_get_from_dump = self.only_get_from_dump_
        )

    def add_height_filtrations_(self, binarizer_threshold: float, *shape: typing.List[int]):
        directions = []
        if self.settings_.height_directions is None:
            directions = list(itertools.product(*([ [ -1, 0, 1 ] ] * len(shape))))
            directions = filter(lambda item: not all(i == 0 for i in item), directions)
            directions = list(directions)
        else:
            directions = self.settings_.height_directions

        for direction in directions:
            self.filtration_extractors_.append((
                self.make_filtration_(gtda.images.HeightFiltration, { 'direction': numpy.array(direction) }, binarizer_threshold),
                f'{int(binarizer_threshold * 10)}/HeightFiltration_{direction[0]}_{direction[1]}',
                f'HeightFiltration with d = ({direction[0]}, {direction[1]}), bin. thr. = 0.{int(binarizer_threshold * 10)}'
            ))
            
    def add_radial_filtrations_(self, binarizer_threshold: float, *shape: typing.List[int]):
        points = [ cvtda.utils.spread_points(coord, self.settings_.num_radial) for coord in shape ]
        for center in list(itertools.product(*points)):
            self.filtration_extractors_.append((
                self.make_filtration_(gtda.images.RadialFiltration, { 'center': numpy.array(center) }, binarizer_threshold),
                f'{int(binarizer_threshold * 10)}/RadialFiltration_{center[0]}_{center[1]}',
                f'RadialFiltration with c = ({center[0]}, {center[1]}), bin. thr. = 0.{int(binarizer_threshold * 10)}'
            ))

    def add_dilation_filtrations_(self, binarizer_threshold: float):
        if not self.settings_.dilation:
            return
        self.filtration_extractors_.append((
            self.make_filtration_(gtda.images.DilationFiltration, { }, binarizer_threshold),
            f'{int(binarizer_threshold * 10)}/DilationFiltration',
            f'DilationFiltration, bin. thr. = 0.{int(binarizer_threshold * 10)}'
        ))

    def add_erosion_filtrations_(self, binarizer_threshold: float):
        if not self.settings_.erosion:
            return
        self.filtration_extractors_.append((
            self.make_filtration_(gtda.images.ErosionFiltration, { }, binarizer_threshold),
            f'{int(binarizer_threshold * 10)}/ErosionFiltration',
            f'ErosionFiltration, bin. thr. = 0.{int(binarizer_threshold * 10)}'
        ))

    def add_signed_distance_filtrations_(self, binarizer_threshold: float):
        if not self.settings_.signed_distance:
            return
        self.filtration_extractors_.append((
            self.make_filtration_(gtda.images.SignedDistanceFiltration, { }, binarizer_threshold),
            f'{int(binarizer_threshold * 10)}/SignedDistanceFiltration',
            f'SignedDistanceFiltration, bin. thr. = 0.{int(binarizer_threshold * 10)}'
        ))
        
    def add_density_filtrations_(self, binarizer_threshold: float):
        for radius in self.settings_.density_radiuses:
            self.filtration_extractors_.append((
                self.make_filtration_(gtda.images.DensityFiltration, { 'radius': radius }, binarizer_threshold),
                f'{int(binarizer_threshold * 10)}/DensityFiltration_{radius}',
                f'DensityFiltration with r = {radius}, bin. thr. = 0.{int(binarizer_threshold * 10)}'
            ))



import numpy
import sklearn.metrics

def estimate_quality(pred_proba: numpy.ndarray, masks: numpy.ndarray) -> dict:
    # Segmentation is basically just per-pixel classification
    scores = pred_proba.flatten()
    preds = (scores >= 0.5).astype(numpy.int8)
    targets = masks.flatten().astype(numpy.int8)
    return {
        'Accuracy':      sklearn.metrics.accuracy_score (targets, preds),
        'AUC-ROC':       sklearn.metrics.roc_auc_score  (targets, scores),
        'Precision':     sklearn.metrics.precision_score(targets, preds, zero_division = 0),
        'Recall':        sklearn.metrics.recall_score   (targets, preds, zero_division = 0),
        'F1-score':      sklearn.metrics.f1_score       (targets, preds, zero_division = 0),
        'Jaccard score': sklearn.metrics.jaccard_score  (targets, preds, zero_division = 0)
    }



import os
import typing

import numpy
import torch
import pandas
import xgboost
import catboost
import torchvision
import sklearn.base
import sklearn.ensemble
import sklearn.neighbors
import matplotlib.pyplot as plt

import cvtda.logging
import cvtda.dumping
import cvtda.neural_network

from .NNClassifier import NNClassifier
from .estimate_quality import estimate_quality

def classify(
    train_images: numpy.ndarray,
    train_features: numpy.ndarray,
    train_labels: numpy.ndarray,
    train_diagrams: typing.Optional[typing.List[numpy.ndarray]],

    test_images: numpy.ndarray,
    test_features: numpy.ndarray,
    test_labels: numpy.ndarray,
    test_diagrams: typing.Optional[typing.List[numpy.ndarray]],

    label_names: typing.Optional[typing.List[str]] = None,
    confusion_matrix_include_values: bool = True,

    n_jobs: int = -1,
    random_state: int = 42,
    dump_name: typing.Optional[str] = None,
    only_get_from_dump: bool = False,

    knn_neighbours: int = 50,

    random_forest_estimators: int = 100,

    nn_device: torch.device = cvtda.neural_network.default_device,
    nn_batch_size: int = 128,
    nn_learning_rate: float = 1e-3,
    nn_epochs: int = 20,
    nn_base = torchvision.models.resnet34,

    grad_boost_max_iter: int = 20,
    grad_boost_max_depth: int = 4,
    grad_boost_max_features: float = 0.1,

    xgboost_n_classifiers: int = 25,
    xgboost_max_depth: int = 4,
    xgboost_device: str = 'gpu',

    catboost_iterations: int = 600,
    catboost_depth: int = 4,
    catboost_device: str = ('GPU' if torch.cuda.is_available() else 'CPU')
):
    without_diagrams = (train_diagrams is None) and (test_diagrams is None)

    if (train_images is not None) and (not only_get_from_dump):
        nn_train = cvtda.neural_network.Dataset(
            train_images, train_diagrams, train_features, train_labels, n_jobs = n_jobs, device = nn_device
        )
        nn_test = cvtda.neural_network.Dataset(
            test_images, test_diagrams, test_features, test_labels, n_jobs = n_jobs, device = nn_device
        )

    def classify_one(classifier: sklearn.base.ClassifierMixin, name: str, display_name: str, ax: plt.Axes):
        if without_diagrams and name == 'NNClassifier_diagrams':
            cvtda.logging.logger().print(f'Skipping {name} - {classifier}')
            return {}

        cvtda.logging.logger().print(f'Trying {name} - {classifier}')

        dumper = cvtda.dumping.dumper()
        model_dump_name = cvtda.dumping.dump_name_concat(dump_name, name)
        if only_get_from_dump or dumper.has_dump(model_dump_name):
            y_pred_proba = dumper.get_dump(model_dump_name)
        else:
            if type(classifier) == NNClassifier:
                classifier.fit(nn_train, nn_test)
                y_pred_proba = classifier.predict_proba(nn_test)
            else:
                classifier.fit(train_features, train_labels)
                y_pred_proba = classifier.predict_proba(test_features)

            if model_dump_name is not None:
                dumper.save_dump(y_pred_proba, model_dump_name)
                
        ax.set_title(display_name)
        result = {
            'classifier': display_name,
            **estimate_quality(
                y_pred_proba, test_labels, ax,
                label_names = label_names, confusion_matrix_include_values = confusion_matrix_include_values
            )
        }
        cvtda.logging.logger().print(result)
        return result

    classifiers = [
        sklearn.neighbors.KNeighborsClassifier(
            n_jobs = n_jobs,
            n_neighbors = knn_neighbours
        ),
        sklearn.ensemble.RandomForestClassifier(
            n_estimators = random_forest_estimators,
            random_state = random_state,
            n_jobs = n_jobs
        ),
        sklearn.ensemble.HistGradientBoostingClassifier(
            random_state = random_state,
            max_iter = grad_boost_max_iter,
            max_depth = grad_boost_max_depth,
            max_features = grad_boost_max_features,
            verbose = cvtda.logging.logger().verbosity()
        ),
        catboost.CatBoostClassifier(
            iterations = catboost_iterations,
            depth = catboost_depth,
            random_seed = random_state,
            loss_function = 'MultiClass',
            devices = '0-3',
            task_type = catboost_device,
            verbose = (cvtda.logging.logger().verbosity() != 0)
        ),
        xgboost.XGBClassifier(
            n_jobs = n_jobs,
            n_estimators = xgboost_n_classifiers,
            max_depth = xgboost_max_depth,
            device = xgboost_device
        ),
        NNClassifier(
            random_state = random_state,
            device = nn_device,
            batch_size = nn_batch_size,
            learning_rate = nn_learning_rate * 10,
            n_epochs = nn_epochs * 2,
            skip_diagrams = True,
            skip_images = True,
            skip_features = False,
            base = nn_base,
        ),
        NNClassifier(
            random_state = random_state,
            device = nn_device,
            batch_size = nn_batch_size,
            learning_rate = nn_learning_rate,
            n_epochs = nn_epochs // 2,
            skip_diagrams = False,
            skip_images = True,
            skip_features = True,
            base = nn_base,
        ),
        NNClassifier(
            random_state = random_state,
            device = nn_device,
            batch_size = nn_batch_size,
            learning_rate = nn_learning_rate,
            n_epochs = nn_epochs,
            skip_diagrams = True,
            skip_images = False,
            skip_features = True,
            base = nn_base,
        ),
        NNClassifier(
            random_state = random_state,
            device = nn_device,
            batch_size = nn_batch_size,
            learning_rate = nn_learning_rate,
            n_epochs = nn_epochs,
            skip_diagrams = True,
            skip_images = False,
            skip_features = False,
            base = nn_base,
        )
    ]
    names = [
        'KNeighborsClassifier',
        'RandomForestClassifier',
        'HistGradientBoostingClassifier',
        'CatBoostClassifier',
        'XGBClassifier',
        'NNClassifier_features',
        'NNClassifier_diagrams',
        'NNClassifier_images',
        'NNClassifier_features_images'
    ]

    display_names = [
        'KNN',
        'Random forest',
        'Histogram-based boosting',
        'CatBoost',
        'XGBoost',
        'FC over topological features',
        'Trainable vectorization',
        'Baseline model',
        'Combined neural network'
    ]

    figure, axes = plt.subplots(3, 3, figsize = (15, 15))
    df = pandas.DataFrame([ classify_one(*args) for args in zip(classifiers, names, display_names, axes.flat) ])
    figure.tight_layout()

    dumper = cvtda.dumping.dumper()
    if (dump_name is not None) and isinstance(dumper, cvtda.dumping.NumpyDumper):
        file = dumper.get_file_name_(cvtda.dumping.dump_name_concat(dump_name, "confusion_matrixes"))
        os.makedirs(os.path.dirname(file), exist_ok = True)
        figure.savefig(file[:-4] + ".svg")
        figure.savefig(file[:-4] + ".png")
        df.to_csv(dumper.get_file_name_(cvtda.dumping.dump_name_concat(dump_name, "quality_metrics.csv"))[:-4])
    return df, figure



import math
import typing
import dataclasses

import numpy
import joblib
import sklearn.base
import gtda.diagrams

import cvtda.utils
import cvtda.logging


def determine_filtering_epsilon_(diagrams: numpy.ndarray, percentile: float) -> float:
    life = (diagrams[:, :, 1] - diagrams[:, :, 0]).flatten()
    if len(numpy.unique(life)) == 1:
        return 1e-8
    return numpy.percentile(life[life != 0], percentile)

class Vectorizer(cvtda.utils.FeatureExtractorBase):
    def __init__(self):
        self.feature_names_ = []
        self.fitted_ = False
    
    def fit(self, diagrams: numpy.ndarray):
        self.fitted_ = True
        self.homology_dimensions_ = numpy.unique(diagrams[:, :, 2])
        num_features = self.transform(diagrams[:32]).shape[1]
        self.feature_names_ = [ f"{i}" for i in range(num_features) ]
        return self

    def transform(self, diagrams: numpy.ndarray) -> numpy.ndarray:
        assert self.fitted_ is True, 'fit() must be called before feature_names()'

    def feature_names(self) -> typing.List[str]:
        assert self.fitted_ is True, 'fit() must be called before feature_names()'
        return self.feature_names_

class SequenceStats(Vectorizer):
    def __init__(
        self,
        impl: typing.List[sklearn.base.TransformerMixin],
        enabled: bool = True,
        reduced_stats: bool = True
    ):
        super().__init__()
        self.impl_ = impl if enabled else []
        self.reduced_stats_ = reduced_stats
    
    def fit(self, diagrams: numpy.ndarray):
        for item in self.impl_:
            item.fit(diagrams)
        return super().fit(diagrams)

    def transform(self, diagrams: numpy.ndarray) -> numpy.ndarray:
        super().transform(diagrams)
        if len(self.impl_) == 0:
            return numpy.empty((len(diagrams), 0))
        flat_shape = (len(diagrams), len(self.homology_dimensions_), -1)
        return numpy.hstack([
            self.calc_stats_(item.transform(diagrams).reshape(flat_shape))
            for item in self.impl_
        ])

    def calc_stats_(self, data: numpy.ndarray):
        return numpy.hstack([
            cvtda.utils.sequence2features(data[:, dim, :], reduced = self.reduced_stats_)
            for dim in range(data.shape[1])
        ])

class Proxy(Vectorizer):
    def __init__(self, impl: sklearn.base.TransformerMixin, enabled: bool = True):
        self.impl_ = impl if enabled else None

    def fit(self, diagrams: numpy.ndarray):
        if self.impl_ is not None:
            self.impl_.fit(diagrams)
        return super().fit(diagrams)

    def transform(self, diagrams: numpy.ndarray) -> numpy.ndarray:
        super().transform(diagrams)
        if self.impl_ is not None:
            return self.impl_.transform(diagrams)
        return numpy.empty((len(diagrams), 0))


class BettiCurve(SequenceStats):
    @dataclasses.dataclass(frozen = True)
    class Settings:
        enabled: bool = True
        n_bins: int = 64
        reduced_stats: bool = True

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full = Settings(reduced_stats = False), reduced = Settings(), quick = Settings()
    )

    def __init__(self, settings = Settings()):
        super().__init__(
            [gtda.diagrams.BettiCurve(n_bins = settings.n_bins, n_jobs = 1)],
            settings.enabled,
            settings.reduced_stats
        )

class Landscape(SequenceStats):
    @dataclasses.dataclass(frozen = True)
    class Settings:
        enabled: bool = True
        n_bins: int = 64
        layers: int = 3
        reduced_stats: bool = True

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full = Settings(reduced_stats = False), reduced = Settings(), quick = Settings()
    )
    
    def __init__(self, settings = Settings()):
        impl = gtda.diagrams.PersistenceLandscape(
            n_layers = settings.layers, n_bins = settings.n_bins, n_jobs = 1
        )
        super().__init__([impl], settings.enabled, settings.reduced_stats)

    def transform(self, diagrams: numpy.ndarray) -> numpy.ndarray:
        assert self.fitted_ is True, 'fit() must be called before feature_names()'
        if len(self.impl_) == 0:
            return numpy.empty((len(diagrams), 0))
        impl: gtda.diagrams.PersistenceLandscape = self.impl_[0]
        landscape = impl.transform(diagrams)
        return numpy.hstack([
            self.calc_stats_(landscape[:, layer::impl.n_layers, :])
            for layer in range(impl.n_layers)
        ])

class Silhouette(SequenceStats):
    @dataclasses.dataclass(frozen = True)
    class Settings:
        enabled: bool = True
        n_bins: int = 64
        powers: typing.List[int] = dataclasses.field(default_factory = lambda: [ 1, 2 ])
        reduced_stats: bool = True

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full = Settings(reduced_stats = False), reduced = Settings(), quick = Settings()
    )
    
    def __init__(self, settings = Settings()):
        impl = [
            gtda.diagrams.Silhouette(power = power, n_bins = settings.n_bins, n_jobs = 1)
            for power in settings.powers
        ]
        super().__init__(impl, settings.enabled, settings.reduced_stats)

class HeatKernel(SequenceStats):
    @dataclasses.dataclass(frozen = True)
    class Settings:
        enabled: bool = True
        n_bins: int = 64
        sigmas: typing.List[float] = dataclasses.field(default_factory = lambda: [ 0.1, 1.0, numpy.pi ])
        reduced_stats: bool = True

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full = Settings(reduced_stats = False), reduced = Settings(enabled = False), quick = Settings(enabled = False)
    )
    
    def __init__(self, settings = Settings()):
        impl = [
            gtda.diagrams.HeatKernel(sigma = sigma, n_bins = settings.n_bins, n_jobs = 1)
            for sigma in settings.sigmas
        ]
        super().__init__(impl, settings.enabled, settings.reduced_stats)

class PersistenceImage(SequenceStats):
    @dataclasses.dataclass(frozen = True)
    class Settings:
        enabled: bool = True
        n_bins: int = 64
        sigmas: typing.List[float] = dataclasses.field(default_factory = lambda: [ 0.1, 1.0, numpy.pi ])
        reduced_stats: bool = True

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full = Settings(reduced_stats = False), reduced = Settings(enabled = False), quick = Settings(enabled = False)
    )
    
    def __init__(self, settings = Settings()):
        impl = [
            gtda.diagrams.PersistenceImage(sigma = sigma, n_bins = settings.n_bins, n_jobs = 1)
            for sigma in settings.sigmas
        ]
        super().__init__(impl, settings.enabled, settings.reduced_stats)

class Entropy(Proxy):
    @dataclasses.dataclass(frozen = True)
    class Settings:
        enabled: bool = True

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(full = Settings(), reduced = Settings(), quick = Settings())
     
    def __init__(self, settings = Settings()):
        super().__init__(
            gtda.diagrams.PersistenceEntropy(nan_fill_value = 0, n_jobs = 1),
            settings.enabled
        )

class NumberOfPoints(Proxy):
    @dataclasses.dataclass(frozen = True)
    class Settings:
        enabled: bool = True

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(full = Settings(), reduced = Settings(), quick = Settings())
    
    def __init__(self, settings = Settings()):
        super().__init__(gtda.diagrams.NumberOfPoints(n_jobs = 1), settings.enabled)

class Lifetime(SequenceStats):
    @dataclasses.dataclass(frozen = True)
    class Settings:
        enabled: bool = True
        reduced_stats: bool = True
        filtering_percentile: float = 10

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full = Settings(reduced_stats = False), reduced = Settings(), quick = Settings()
    )
    
    def __init__(self, settings = Settings()):
        super().__init__([], settings.enabled, settings.reduced_stats)
        self.filtering_percentile_ = settings.filtering_percentile
        self.enabled_ = settings.enabled
    
    def fit(self, diagrams: numpy.ndarray):
        self.filtering_epsilon_ = determine_filtering_epsilon_(diagrams, self.filtering_percentile_)
        return super().fit(diagrams)

    def transform(self, diagrams: numpy.ndarray) -> numpy.ndarray:
        Vectorizer.transform(self, diagrams)
        if not self.enabled_:
            return numpy.empty((len(diagrams), 0))
        
        birth, death, dim = diagrams[:, :, 0], diagrams[:, :, 1], diagrams[:, :, 2]
        bd2, life = (birth + death) / 2.0, death - birth
        bd2_bulk, life_bulk = [ ], [ ]
        for d in self.homology_dimensions_:
            mask = (dim != d) | (life < self.filtering_epsilon_)
            bd2_bulk.append(numpy.ma.array(bd2, mask = mask))
            life_bulk.append(numpy.ma.array(life, mask = mask))
        return numpy.hstack([
            self.calc_stats_(numpy.ma.stack(bd2_bulk, axis = 1)),
            self.calc_stats_(numpy.ma.stack(life_bulk, axis = 1))
        ])

class DiagramVectorizer(cvtda.utils.FeatureExtractorBase):
    @dataclasses.dataclass(frozen = True)
    class Settings:
        filtering_percentile: float                     = 10
        betti_curve:          BettiCurve.Settings       = BettiCurve.Settings()
        landscape:            Landscape.Settings        = Landscape.Settings()
        silhouette:           Silhouette.Settings       = Silhouette.Settings()
        heat_kernel:          HeatKernel.Settings       = HeatKernel.Settings()
        persistence_image:    PersistenceImage.Settings = PersistenceImage.Settings()
        entropy:              Entropy.Settings          = Entropy.Settings()
        number_of_points:     NumberOfPoints.Settings   = NumberOfPoints.Settings()
        lifetime:             Lifetime.Settings         = Lifetime.Settings()

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full = Settings(
            betti_curve       = BettiCurve.      PRESETS.full,
            landscape         = Landscape.       PRESETS.full,
            silhouette        = Silhouette.      PRESETS.full,
            heat_kernel       = HeatKernel.      PRESETS.full,
            persistence_image = PersistenceImage.PRESETS.full,
            entropy           = Entropy.         PRESETS.full,
            number_of_points  = NumberOfPoints.  PRESETS.full,
            lifetime          = Lifetime.        PRESETS.full
        ),
        reduced = Settings(
            betti_curve       = BettiCurve.      PRESETS.reduced,
            landscape         = Landscape.       PRESETS.reduced,
            silhouette        = Silhouette.      PRESETS.reduced,
            heat_kernel       = HeatKernel.      PRESETS.reduced,
            persistence_image = PersistenceImage.PRESETS.reduced,
            entropy           = Entropy.         PRESETS.reduced,
            number_of_points  = NumberOfPoints.  PRESETS.reduced,
            lifetime          = Lifetime.        PRESETS.reduced
        ),
        quick = Settings(
            betti_curve       = BettiCurve.      PRESETS.quick,
            landscape         = Landscape.       PRESETS.quick,
            silhouette        = Silhouette.      PRESETS.quick,
            heat_kernel       = HeatKernel.      PRESETS.quick,
            persistence_image = PersistenceImage.PRESETS.quick,
            entropy           = Entropy.         PRESETS.quick,
            number_of_points  = NumberOfPoints.  PRESETS.quick,
            lifetime          = Lifetime.        PRESETS.quick
        )
    )

    EXTRACTOR_NAMES = [ "betti", "landscape", "silhouette", "heat", "persistence_image", "entropy", "number_of_points", "lifetime" ]

    def __init__(self, n_jobs: int = -1, batch_size: int = None, settings: Settings = PRESETS.reduced):
        self.fitted_ = False
        self.n_jobs_ = n_jobs
        self.batch_size_ = batch_size

        self.filtering_percentile_ = settings.filtering_percentile
        self.extractors_: typing.List[cvtda.utils.FeatureExtractorBase] = [
            BettiCurve(settings.betti_curve),
            Landscape(settings.landscape),
            Silhouette(settings.silhouette),
            HeatKernel(settings.heat_kernel),
            PersistenceImage(settings.persistence_image),
            Entropy(settings.entropy),
            NumberOfPoints(settings.number_of_points),
            Lifetime(settings.lifetime)
        ]

    def feature_names(self) -> typing.List[str]:
        assert self.fitted_ is True, 'fit() must be called before feature_names()'
        feature_names = []
        for extractor, name in zip(self.extractors_, DiagramVectorizer.EXTRACTOR_NAMES):
            feature_names.extend(self.nest_feature_names(name, extractor.feature_names()))
        return feature_names

    def fit(self, diagrams: numpy.ndarray):
        self.filtering_epsilon_ = determine_filtering_epsilon_(diagrams, self.filtering_percentile_)
        self.filtering_ = gtda.diagrams.Filtering(epsilon = self.filtering_epsilon_).fit(diagrams)
        diagrams = self.filtering_.transform(diagrams)
        for extractor in self.extractors_:
            extractor.fit(diagrams)
        cvtda.logging.logger().print('DiagramVectorizer: fitting complete')
        self.fitted_ = True
        return self
    
    def transform(self, diagrams: numpy.ndarray) -> numpy.ndarray:
        assert self.fitted_ is True, 'fit() must be called before transform()'
        batch_size = self.get_batch_size_(len(diagrams))
        loop = list(range(0, len(diagrams), batch_size))
        
        def transform_batch(batch_start: numpy.ndarray) -> numpy.ndarray:
            return self.transform_batch_raw_(diagrams[batch_start:batch_start + batch_size])
        features = cvtda.utils.parallel(transform_batch, loop, return_as = 'generator', n_jobs = self.n_jobs_)
        collector = cvtda.logging.logger().pbar(features, total = len(loop), desc = 'DiagramVectorizer: batch')
        features = numpy.vstack(list(collector))
        assert features.shape == (len(diagrams), len(self.feature_names())), \
                f"{features.shape} != {(len(diagrams), len(self.feature_names()))}"
        return features

    def get_batch_size_(self, num_objects: int):
        if self.batch_size_ is not None:
            return self.batch_size_
        n_jobs = joblib.effective_n_jobs(self.n_jobs_)
        batch_size = math.ceil(num_objects / n_jobs)
        return math.ceil(batch_size / math.ceil(batch_size / 256))

    def transform_batch_raw_(self, batch: numpy.ndarray) -> numpy.ndarray:
        batch = self.filtering_.transform(batch)
        return numpy.hstack([ extractor.transform(batch) for extractor in self.extractors_ ])



import torch

from .downsample import DownsampleBlock
from .upsample import UpsampleBlock
from .bottom import Bottom

class MiniUnetModule(torch.nn.Module):
    def __init__(
        self,
        images_example: torch.Tensor,
        features_example: torch.Tensor,
        with_images: bool = True,
        with_features: bool = True,
        layers: int = 4,
        start_channels: int = 8,
        remove_cross_maps: bool = False
    ):
        super().__init__()
        self.remove_cross_maps_ = remove_cross_maps
        start_channels = max(images_example.shape[1], start_channels)
        channels = [ images_example.shape[1], *[ (2 ** i) * start_channels for i in range(layers + 1) ] ]

        self.downsample_blocks = torch.nn.ModuleList([ DownsampleBlock(*c) for c in zip(channels[:-2], channels[1:-1]) ])
        for downsample_block in self.downsample_blocks:
            _, images_example = downsample_block(images_example)

        self.bottom = Bottom(images_example, features_example, with_images, with_features, channels[-2], channels[-1])
        
        channels.reverse()
        self.upsample_blocks = torch.nn.ModuleList([ UpsampleBlock(*c) for c in zip(channels[:-2], channels[1:-1]) ])
        self.output = torch.nn.Sequential(torch.nn.Conv2d(channels[-2], 1, kernel_size = 1), torch.nn.Sigmoid())

    def forward(self, images: torch.Tensor, features: torch.Tensor):
        feature_maps = [ ]
        for downsample_block in self.downsample_blocks:
            feature_map, images = downsample_block(images)
            feature_maps.insert(0, feature_map)
            
        images = self.bottom(images, features)
        
        for upsample_block, feature_map in zip(self.upsample_blocks, feature_maps):
            if self.remove_cross_maps_:
                feature_map = torch.zeros_like(feature_map)
            images = upsample_block(images, feature_map)
        return self.output(images)
    
    def num_params(self) -> int:
        return sum(p.numel() for p in self.parameters())



import numpy

def aggregate(diff: numpy.ndarray) -> float:
    axes = tuple(set(range(len(diff.shape))) - set([0]))
    return numpy.sum(diff, axis = axes).mean()


def estimate_quality(decoded: numpy.ndarray, original: numpy.ndarray) -> dict:
    decoded = numpy.squeeze(decoded)
    original = numpy.squeeze(original)
    return {
        'MAE': aggregate(numpy.abs(original - decoded)),
        'MSE': aggregate((original - decoded) ** 2)
    }
